package project1;

import java.awt.Color;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import jv.geom.PgElementSet;
import jv.geom.PgPointSet;
import jv.geom.PgPolygonSet;
import jv.loader.PjImportModel;
import jv.number.PuDouble;
import jv.number.PuInteger;
import jv.object.PsConfig;
import jv.object.PsDebug;
import jv.project.PgGeometryIf;
import jv.project.PjProject;
import jv.project.PvViewerIf;
import jv.vecmath.PdVector;
import jv.vecmath.PiVector;
import jvx.curve.PgTube;
import jvx.geom.PwRefineElementSet;
import jvx.geom.PwRefinePolygon;
import jvx.geom.PwSimplify;
import jvx.numeric.PnPCA;

public class PjProject1 extends PjProject implements ActionListener {
	private static final long serialVersionUID = 1L;

	/* Loader */
	static String m_defFileName = "models\\cube.obj";
	protected String m_fileName;
	protected PjImportModel m_import;

	/**
	 * Constructor, without arguments to allow loading of project from menu.
	 */
	public PjProject1() {
		this(PsConfig.getCodeBase() + m_defFileName);
	}

	public PjProject1(String fileName) {
		super(fileName);
		if (fileName == null)
			fileName = "Unknown file name";
		m_fileName = fileName;

		m_import = new PjImportModel();
		m_import.setTypeOfInfoPanel(PjImportModel.SMALL_INFO_PANEL);
		m_import.setFileName(fileName);
		m_import.addActionListener(this);

		if (getClass() == PjProject1.class) {
			init();
		}
	}

	/*
	 * 
	 * 
	 * PgPointSet Section
	 * 
	 * 
	 */
	// Global Parameter
	PuDouble m_radius;

	/**
	 * Generates for one marked vertex the geometric neighborhood for a given radius
	 * and colors all points within range in blue, otherwise white.
	 */
	public void pointSet_Neighborhood() {
		resetGeometry();

		// Call loaded geometry to points and get its number of points
		PgPointSet points = (PgPointSet) m_geom;
		int nov = points.getNumVertices();

		// Check if one vertex is marked
		if (points.getMarkedVertices().length == 1) {

			// Collect neighboring indices
			PiVector neighbors = generateNeighborhood(points, points.getMarkedVertices()[0], m_radius.getValue());

			// Color neighbored in blue, otherwise white
			for (int i = 0; i < nov; i++) {
				if (neighbors.contains(i))
					points.setVertexColor(i, Color.BLUE);
				else
					points.setVertexColor(i, Color.WHITE);
			}

			// Show colors and update
			points.showVertexColors(true);
			m_display.update(points);
		}

		// If none ore more than one is selected
		else
			PsDebug.showStatus("Warning: Either no or more than one point selected.");
		
	}

	/**
	 * Generates for every point a geometric neighborhood for a given radius, and
	 * sets an non-oriented normal at every vertex obtained by an extra class
	 * involving principal component analysis.
	 */
	public void pointSet_NormalEstimation() {
		resetGeometry();

		// Call loaded geometry and its number of points
		PgElementSet points = (PgElementSet) m_geom;
		int nov = points.getNumVertices();

		// Access every point
		for (int i = 0; i < nov; i++) {

			// Collect the points' neighbors
			PiVector neighbors = generateNeighborhood(points, i, m_radius.getValue());

			// Create barycenter (mean) of neighboring points
			PdVector barycenter = new PdVector(3);
			for (int j : neighbors.m_data)
				barycenter.add(points.getVertex(j));
			barycenter.multScalar(1. / neighbors.getSize());

			// Set vertex normal according to PCA (normal to least squares plane)
			points.setVertexNormal(i, PnPCA.getNormal(points, barycenter, neighbors.m_data));
		}

		// Show normals and color + update
		points.showVertexNormals(true);
		points.setGlobalVertexNormalColor(Color.BLUE);
		m_display.update(points);
	}

	/**
	 * Generates the geometric neighborhood for a given vertex index and radius, and
	 * returns a vector of vertex indices lying within range.
	 * 
	 * @param points     The underlying point set for queries.
	 * @param pointIndex The point index counting as center of the radial ball.
	 * @param radius     The radius for the radial ball.
	 * @return A vector of indices lying within range.
	 */
	PiVector generateNeighborhood(PgPointSet points, int pointIndex, double radius) {
		// Instantiate neighbor array and copy point for given index
		PiVector neighbors = new PiVector();
		PdVector point = PdVector.copyNew(points.getVertex(pointIndex));
		int nov = points.getNumVertices();

		// Pairwise check, whether Euclidean distance is smaller equal to radius; if so add as neighbor
		for (int i = 0; i < nov; i++)
			if (PdVector.dist(points.getVertex(i), point) <= radius)
				neighbors.addEntry(i);

		return neighbors;
	}

	/*
	 * 
	 * 
	 * PgPolygonSet Section
	 * 
	 * 
	 */
	// Global Parameter
	protected PuInteger m_refinementScale;

	/**
	 * Generates a polygon set out of the first eight vertices of the loaded
	 * geometry and closes it. Further it returns it.
	 * 
	 * @return The generated polygon set.
	 */
	public PgPolygonSet polygonSet_generate() {
		resetGeometry();

		// Generate polygonset from first eight vertices of loaded geometry plus polygon connecting all
		PgPolygonSet polygons = new PgPolygonSet(3);
		polygons.setName("PolygonSet");
		PiVector polygon = new PiVector(8);
		for (byte i = 0; i < 8; i++) {
			polygons.addVertex(m_geom.getVertex(i));
			polygon.setEntry(i, i);
		}
		polygons.addPolygon(polygon);
		polygons.setClosed(0, true);

		// Add geometry and update; further return it
		m_display.addGeometry(polygons);
		m_display.update(polygons);
		return polygons;
	}

	/**
	 * Generates a polygon set (from method: polygonSet_generate) and builds tubes
	 * around its edges.
	 */
	public void polygonSet_Tubes() {
		resetGeometry();

		// Call polygon generated by method
		PgPolygonSet polygons = polygonSet_generate();

		// Generate tube structure and pass polygon
		PgTube tubes = new PgTube(3);
		tubes.setName("Tubes");
		tubes.setPolygonSet(polygons);
		tubes.setDiscr(20);
		tubes.setThickness(.2);
		tubes.computeTube();

		// Add tube to display and update
		m_display.addGeometry(tubes);
		m_display.update(tubes);
	}

	/**
	 * Generates a polygon set (from method: polygonSet_generate) and uses a slider
	 * to refine the polygon set w.r.t. some refinement algorithm.
	 */
	public void polygonSet_Refinement() {
		resetGeometry();

		// Call polygon generated by method
		PgPolygonSet polygons = polygonSet_generate();

		// Take scale value and refine several times
		int scale = m_refinementScale.getValue();
		for (int i = 0; i < scale; i++)
			PwRefinePolygon.refineChaikin(polygons);

		// Update polygon
		m_display.update(polygons);
	}

	/*
	 * 
	 * 
	 * PgElementSet Section
	 * 
	 * 
	 */
	// Parameters
	PiVector[] neighborElements;

	/**
	 * Copies the loaded geometry and either simplifies or subdivides, according to
	 * the choice provided. Simplification reduces the number of elements down to
	 * one third. The subdivision refines the element set w.r.t. some refinement
	 * algorithm.
	 * 
	 * @param choice 0 - Simplification; 1 - Subdivision
	 */
	public void elementSet_SimplificationSubdivision(int choice) {
		resetGeometry();

		// Copy loaded geometry to element set
		PgElementSet elements = new PgElementSet(3);
		elements.copy((PgElementSet) m_geom);
		elements.setName("ElementSet");

		// If choice 0 apply simplification via workshop and reduce number of elements down to 1/3
		if (choice == 0) {
			// 
			int noe = elements.getNumElements();
			PwSimplify simplification = new PwSimplify();
			simplification.setGeometry(elements);
			simplification.simplify(noe / 3);
		}

		// Otherwise refine due to sqrt2
		else
			PwRefineElementSet.refineSqrt2(elements);

		// Add geometry and update
		m_display.addGeometry(elements);
		m_display.update(elements);
	}

	/**
	 * Identifies marked elements, generates vectors of neighboring elements, and
	 * stores them in the global array of vectors in variable: neighborElements.
	 * Afterwards it colors the collected elements in green, otherwise white.
	 * 
	 * @return The colored element set.
	 */
	public PgElementSet elementSet_identifyNeighbors() {
		resetGeometry();

		// Call lodaed geometry as element set and its number of points
		PgElementSet elements = (PgElementSet) m_geom;
		int noe = elements.getNumElements();

		// Set up vector of marked elements and store these which are tagged as selected
		PiVector markedElements = new PiVector();
		for (int i = 0; i < noe; i++)
			if (elements.getElement(i).hasTag(IS_SELECTED))
				markedElements.addEntry(i);
		PsDebug.showStatus("Marked Element indices are: " + markedElements.toShortString());

		// If there are marked elements
		if (markedElements.getSize() != 0) {

			// Get number of marked elements and once store for every its neighboring elements; also store all globally for coloring
			int none = markedElements.getSize();
			neighborElements = new PiVector[none];
			PiVector allNeighbors = new PiVector();
			for (int j = 0; j < none; j++) {
				neighborElements[j] = PiVector.copyNew(elements.getNeighbour(markedElements.getEntry(j)));
				allNeighbors.concat(neighborElements[j]);
			}

			// Sort global list of neighbor element indices and remove duplicates; color these green; othwerise white
			allNeighbors.sort();
			allNeighbors.removeSuccessiveDuplicates();
			for (int i = 0; i < noe; i++)
				if (allNeighbors.contains(i))
					elements.setElementColor(i, Color.GREEN);
				else
					elements.setElementColor(i, Color.WHITE);

			// Show element colors and update
			elements.showElementColors(true);
			m_display.update(elements);
			return elements;
		}

		// If no elements are marked return null reference
		else
			return null;
	}

	/**
	 * Takes the colored element set (from method: elementSet_identifyNeighbors) and
	 * the neighbor-element relation in variable: neighborElements. For every
	 * neighborhood it sends the vertices contributing vertices to their barycenter.
	 */
	public void elementSet_simplifyNeighbors() {
		resetGeometry();

		// Generate new element set and copy instance from method elementSet_identifyNeighbors()
		PgElementSet elements = new PgElementSet(3);
		elements.setName("ElementSet");
		elements.copy(elementSet_identifyNeighbors());

		// If it has elements (it is not null)
		if (elements.getNumElements() != 0) {

			// Go through every neighor array of marked element and collect point indices
			for (PiVector neighborElement : neighborElements) {
				PiVector vertexIndices = new PiVector();
				for (int elementIndex : neighborElement.m_data)
					vertexIndices.concat(elements.getElement(elementIndex));

				// Sort and remove duplicates of contributint vertex indices
				vertexIndices.sort();
				vertexIndices.removeSuccessiveDuplicates();

				// Calculate barycenter (mean) of remaining indices
				PdVector barycenter = new PdVector(3);
				for (int vertexIndex : vertexIndices.m_data)
					barycenter.add(elements.getVertex(vertexIndex));
				barycenter.multScalar(1. / vertexIndices.getSize());

				// Send all vertices of neighboring elements to calculated barycenter
				for (int vertexIndex : vertexIndices.m_data)
					elements.setVertex(vertexIndex, barycenter);
			}

			// Add new geometry and update
			m_display.addGeometry(elements);
			m_display.update(elements);
		}

		// If geometry had no elements (was null reference; no elements marked)
		else
			PsDebug.showStatus("No elements marked.");
	}

	/*
	 * 
	 * 
	 * Misc
	 * 
	 * 
	 */
	public void method() {

	}

	/*
	 * 
	 * 
	 * Setter Method
	 * 
	 * 
	 */
	/**
	 * Sets visibility options. The choices refer to showing 0-geometry, 1-vertices,
	 * 2-edges, 3-elements of the loaded geometry. The state is obtained by the
	 * checkbox state in the info panel.
	 * 
	 * @param choice Visibility of 0-geometry, 1-vertices, 2-edges, 3-elements.
	 * @param state  Whether to show or not.
	 */
	void setVisibility(int choice, boolean state) {
		PgElementSet g = (PgElementSet) m_geom;
		switch (choice) {
		case 0: // Show geometry
			g.setVisible(state);
			break;
		case 1: // Show vertices
			g.showVertices(state);
			break;
		case 2: // Show edges
			g.showEdges(state);
			break;
		case 3: // Show elements
			g.showElements(state);
			break;
		}
		m_display.update(g);
	}

	/**
	 * Removes all geometries and resets loaded geometry.
	 */
	void resetGeometry() {
		PgElementSet g = (PgElementSet) m_geom;
		g.showVertexColors(false);
		g.showEdgeColors(false);
		g.showElementColors(false);
		g.removeVertexNormals();
		g.removeElementNormals();
		m_display.removeGeometries();
		m_display.addGeometry(m_geom);
		m_display.selectGeometry(m_geom);
		m_display.update(m_geom);
	}

	/*
	 * 
	 * 
	 * Project Initialization and Setup
	 * 
	 * 
	 */
	/**
	 * Do initialization of data structures; method is also used to reset instances.
	 */
	public void init() {
		super.init();
		m_refinementScale = new PuInteger("Refine");
		m_refinementScale.addUpdateListener(this);
		m_refinementScale.setDefBounds(0, 10, 1, 1);
		m_refinementScale.setDefValue(1);
		m_refinementScale.init();

		m_radius = new PuDouble("radius");
		m_radius.setDefBounds(0, 10, 0.1, 1);
		m_radius.setDefValue(1.);
		m_radius.init();
	}

	/**
	 * Start project, e.g. start an animation. Method is called once when project is
	 * selected in PvViewer#selectProject(). Method is optional. For example, if an
	 * applet calls the start() method of PvViewer, then PvViewer tries to invoke
	 * the start() method of the currently selected project.
	 */
	public void start() {
		super.start();
		PvViewerIf viewer = getViewer();
		m_display.setBackgroundColor(Color.WHITE);
		if (viewer != null) {
			String model = viewer.getParameter("model");
			if (model != null) {
				m_fileName = PsConfig.getCodeBase() + model;
			}
		}
		m_import.load(m_fileName);
		if (m_import.getGeometry() instanceof PgElementSet || m_import.getGeometry() instanceof PgPointSet
				|| m_import.getGeometry() instanceof PgPolygonSet) {
			addGeometry(m_import.getGeometry());
			fitDisplays();
		}
		afterGeomLoaded();
	}

	public boolean setGeometry(PgGeometryIf geom) {
		m_geom = geom;
		return true;
	}

	/** Catch action events from the import loader when a new model was loaded. */
	public void actionPerformed(ActionEvent event) {
		if (event.getSource() == m_import) {
			if (PsDebug.NOTIFY)
				PsDebug.notify("(m_import): called");
			if (m_import.getConfirm() == PjImportModel.CONFIRM_CLICKED) {
				if (m_import.getGeometry() instanceof PgElementSet || m_import.getGeometry() instanceof PgPointSet
						|| m_import.getGeometry() instanceof PgPolygonSet) {
					addGeometry(m_import.getGeometry());
					fitDisplays();
				}
			} else if (m_import.getConfirm() == PjImportModel.CONFIRM_CANCEL) {
				fitDisplays();
			} else if (m_import.getConfirm() == PjImportModel.CONFIRM_OK) {
				m_geom = (PgPointSet) m_import.getGeometry();
				if (!addGeometry(m_geom))
					return;
			}
			fitDisplays();
		}
		afterGeomLoaded();
	}

	/**
	 * Called in start(), actionPerformed() and load().
	 */
	public void afterGeomLoaded() {
		System.out.println("Loaded geometry: " + m_fileName);
		m_geom = m_import.getGeometry();
	}

	/**
	 * 
	 * @param filepath
	 */
	public void load(String filepath) {
		m_fileName = filepath;
		m_import.load(m_fileName);
		if (m_import.getGeometry() instanceof PgElementSet || m_import.getGeometry() instanceof PgPointSet
				|| m_import.getGeometry() instanceof PgPolygonSet) {
			addGeometry(m_import.getGeometry());
			fitDisplays();
		}
		afterGeomLoaded();
	}

	/**
	 * Update method of project to react on changes in its panel or of its children.
	 * This method is optional, but required if project is parent of a child.
	 * Project becomes parent if child calls <code>child.setParent(this)</code> with
	 * this project as argument. For example, see the constructor of MyProject.
	 * Project must react on child events, or forward them to its superclass.
	 * <p>
	 * Catch events of integer children and recompute surface.
	 */
	public boolean update(Object event) {
		if (event == m_refinementScale) {
			polygonSet_Refinement();
			return true;
		}

		return super.update(event);
	}
}
